import logging
from collections import OrderedDict
from contextlib import contextmanager

from stevedore import ExtensionManager

from medikit.pipeline import Pipeline

logger = logging.getLogger(__name__)


class ConfigurationRegistry():
    def __init__(self):
        self._configs = {}
        self._features = {}
        self._pipelines = {}
        self._variables = OrderedDict()

        def register_feature(ext):
            self._features[ext.name] = ext.plugin

        def on_load_feature_failure(mgr, entrypoint, err):
            logger.exception('Exception caught while loading {}.'.format(entrypoint), err)

        mgr = ExtensionManager(namespace='medikit.feature', on_load_failure_callback=on_load_feature_failure)
        mgr.map(register_feature)

    def __getitem__(self, item):
        return self._configs[item]

    def __contains__(self, item):
        return item in self._configs

    def set_vars(self, **variables):
        self._variables.update(variables)

    def get_var(self, name, default=None):
        return self._variables.get(name, default)

    @property
    def variables(self):
        return self._variables

    @property
    def package_name(self):
        if 'python' in self:
            return self['python'].get('name')
        else:
            name = self.get_var('PACKAGE')
            if not name:
                raise RuntimeError('You must define a package name, using either python.setup() or PACKAGE = ...')
            return name

    def keys(self):
        return self._configs.keys()

    def require(self, *args):
        if len(args) == 1:
            return self._require(args[0])
        elif len(args) > 1:
            return tuple(map(self._require, args))
        raise ValueError('Empty.')

    @contextmanager
    def pipeline(self, name):
        if not name in self._pipelines:
            self._pipelines[name] = Pipeline()
        yield self._pipelines[name]

    @property
    def pipelines(self):
        return self._pipelines

    def _require(self, name):
        if not name in self._features:
            raise ValueError('Unknown feature {!r}.'.format(name))

        if name not in self._configs:
            self._configs[name] = self._features[name].Config()

        return self._configs[name]



from whistle import Event


class MakefileEvent(Event):
    def __init__(self, package_name, makefile, config):
        self.package_name = package_name
        self.makefile = makefile
        self.config = config
        super(MakefileEvent, self).__init__()


"""
Setup make targets to rollout and rollback this project as a deployment onto a Kubernetes cluster.

"""
import json

from medikit.events import subscribe
from medikit.feature import Feature
from medikit.feature.make import which


class KubeConfig(Feature.Config):
    def __init__(self):
        self._targets = list()
        self._targets_data = dict()

    def add_target(self, name, *, patch):
        if name in self._targets:
            raise ValueError('Kubernetes target {} already defined.'.format(name))
        self._targets.append(name)
        self._targets_data[name] = patch

    def get_targets(self):
        for target in self._targets:
            yield target, self._targets_data[target]


class KubeFeature(Feature):
    Config = KubeConfig

    requires = {'docker'}

    @subscribe('medikit.feature.make.on_generate', priority=-1)
    def on_make_generate(self, event):
        kube_config = event.config['kube']

        event.makefile['KUBECTL'] = which('kubectl')
        event.makefile['KUBECTL_OPTIONS'] = ''
        event.makefile['KUBECONFIG'] = ''
        event.makefile['KUBE_NAMESPACE'] = 'default'

        targets = list(kube_config.get_targets())
        if len(targets):
            event.makefile.add_target(
                'kube-rollout',
                '\n'.join(
                    [
                        '$(KUBECTL) $(KUBECTL_OPTIONS) --namespace=$(KUBE_NAMESPACE) patch {target} -p{patch}'.format(
                            target=target,
                            patch=repr(json.dumps(patch)),
                        ) for target, patch in targets
                    ]
                ),
                phony=True,
                doc='Rollout docker image onto kubernetes cluster.'
            )

            event.makefile.add_target(
                'kube-rollback',
                '\n'.join(['$(KUBECTL) rollout undo {target}'.format(target=target) for target, patch in targets]),
                phony=True,
                doc='Rollbacks last kubernetes patch operation.'
            )



"""
The “python” feature is the base feature required for all python projects.

Medikit only supports python 3.5+ projects, which we believe is a future proof choice.

Overview
::::::::

.. code-block:: python

    from medikit import require

    # load python feature (idempotent).
    python = require('python')

    # configure our package.
    python.setup(
        name = 'awesome-library',
        author = 'Chuck Norris',
    )

    # add base and extras requirements, with "loose" versionning (the frozen version fits better in requirements*.txt)
    python.add_requirements(
        'django',
        dev=[
            'pytest',
        ],
    )

"""

import itertools
import os
import tempfile
from getpass import getuser

from piptools.repositories import PyPIRepository
from piptools.resolver import Resolver
from piptools.scripts.compile import get_pip_command
from piptools.utils import format_requirement
from piptools._vendored.pip.req import parse_requirements
from piptools._vendored.pip._vendor.distlib.util import parse_requirement

from medikit.events import subscribe
from medikit.feature import Feature, ABSOLUTE_PRIORITY
from medikit.feature.make import which
from medikit.utils import get_override_warning_banner


def _normalize_requirement(req):
    bits = req.requirement.split()
    if req.extras:
        bits = [bits[0] + '[{}]'.format(','.join(req.extras))] + bits[1:]
    return ' '.join(bits)


class PythonConfig(Feature.Config):
    """ Configuration API for the «python» feature. """

    on_generate = __name__ + '.on_generate'

    def __init__(self):
        self._setup = {}
        self._requirements = {None: {}, 'dev': {}}
        self._constraints = {}
        self._version_file = None
        self._create_packages = True
        self.override_requirements = False

    @property
    def package_dir(self):
        return '/'.join(self.get('name').split('.'))

    @property
    def version_file(self):
        return self._version_file or os.path.join(self.package_dir, '_version.py')

    @version_file.setter
    def version_file(self, value):
        self._version_file = value

    @property
    def create_packages(self):
        return self._create_packages

    @create_packages.setter
    def create_packages(self, value):
        self._create_packages = bool(value)

    def get(self, item):
        if item == 'install_requires':
            return list(self.get_requirements(with_constraints=True))
        if item == 'extras_require':
            return {
                extra: list(self.get_requirements(extra=extra, with_constraints=True))
                for extra in self.get_extras()
            }
        return self._setup.get(item)

    def get_init_files(self):
        # Explode package name so we know which python packages are namespaced and
        # which are not.
        package_bits = self.get('name').split('.')
        if len(package_bits) > 1:
            # XXX SIDE EFFECT !!!
            self['namespace_packages'] = []
            for i in range(1, len(package_bits)):
                self['namespace_packages'].append('.'.join(package_bits[0:i]))

        for i in range(1, len(package_bits) + 1):
            _bits = package_bits[0:i]
            package_dir = os.path.join(*_bits)
            package_init_file = os.path.join(package_dir, '__init__.py')
            is_namespace = i < len(package_bits)
            yield package_dir, package_init_file, {'is_namespace': is_namespace}

    def __getitem__(self, item):
        return self.get(item)

    def __setitem__(self, key, value):
        self.setup(**{key: value})

    def __contains__(self, item):
        return bool(self.get(item))

    def add_constraints(self, *reqs, **kwargs):
        self.__add_constraints(reqs)
        for extra, reqs in kwargs.items():
            extra = extra.replace('_', '-')
            self.__add_constraints(reqs, extra=extra)
        return self

    def add_requirements(self, *reqs, **kwargs):
        self.__add_requirements(reqs)
        for extra, reqs in kwargs.items():
            extra = extra.replace('_', '-')
            self.__add_requirements(reqs, extra=extra)
        return self

    def get_extras(self):
        return sorted(filter(None, self._requirements.keys()))

    def get_constraints(self, extra=None):
        for name, req in sorted(self._constraints.get(extra, {}).items()):
            yield _normalize_requirement(req)

    def get_requirements(self, extra=None, with_constraints=False):
        if with_constraints:
            yield from self.get_constraints(extra=extra)
        for name, req in sorted(self._requirements[extra].items()):
            yield _normalize_requirement(req)

    def setup(self, **kwargs):
        self._setup.update(kwargs)
        return self

    def get_setup(self):
        return self._setup

    def __add_constraints(self, reqs, extra=None):
        if len(reqs):
            if extra not in self._constraints:
                self._constraints[extra] = {}
        for req in reqs:
            req = parse_requirement(req)
            if req.name in self._constraints[extra]:
                raise ValueError('Duplicate constraint for {}.'.format(req.name))
            self._constraints[extra][req.name] = req

    def __add_requirements(self, reqs, extra=None):
        if len(reqs):
            if extra not in self._requirements:
                self._requirements[extra] = {}
        for req in reqs:
            req = parse_requirement(req)
            if req.name in self._requirements[extra]:
                raise ValueError('Duplicate requirement for {}.'.format(req.name))
            self._requirements[extra][req.name] = req


class PythonFeature(Feature):
    """
    Adds the requirements/requirements-dev logic, virtualenv support, setup.py generation, a few metadata files used by
    setuptools...

    """

    requires = {'make'}

    Config = PythonConfig

    @subscribe('medikit.feature.make.on_generate', priority=ABSOLUTE_PRIORITY)
    def on_make_generate(self, event):
        """
        **Environment variables**

        - ``PACKAGE``
        - ``PYTHON``
        - ``PYTHON_BASENAME``
        - ``PYTHON_DIR``
        - ``PYTHON_REQUIREMENTS_FILE``
        - ``PYTHON_REQUIREMENTS_DEV_FILE``
        
        **Shortcuts**
        - ``PIP``
        - ``PIP_INSTALL_OPTIONS``

        **Make targets**

        - ``install``
        - ``install-dev``

        :param MakefileEvent event:

        """

        extra_variables = []
        for extra in event.config['python'].get_extras():
            extra_variables.append(
                (
                    'PYTHON_REQUIREMENTS_' + extra.upper().replace('-', '_') + '_FILE',
                    'requirements-' + extra + '.txt',
                )
            )

        # Python related environment
        event.makefile.updateleft(
            (
                'PACKAGE',
                event.config.package_name,
            ), (
                'PYTHON',
                which('python'),
            ), (
                'PYTHON_BASENAME',
                '$(shell basename $(PYTHON))',
            ), (
                'PYTHON_DIRNAME',
                '$(shell dirname $(PYTHON))',
            ), (
                'PYTHON_REQUIREMENTS_FILE',
                'requirements.txt',
            ), *extra_variables
        )

        event.makefile['PIP'] = '$(PYTHON) -m pip'
        event.makefile['PIP_INSTALL_OPTIONS'] = ''

        event.makefile.get_target('install').install = [
            '$(PIP) install $(PIP_INSTALL_OPTIONS) -U "pip ~=10.0" wheel',
            '$(PIP) install $(PIP_INSTALL_OPTIONS) -U -r $(PYTHON_REQUIREMENTS_FILE)',
        ]
        event.makefile.get_target('install').deps += ['$(PYTHON_REQUIREMENTS_FILE)', 'setup.py']

        for extra in event.config['python'].get_extras():
            target = event.makefile.add_install_target(extra)
            reqs_var = 'PYTHON_REQUIREMENTS_' + extra.upper().replace('-', '_') + '_FILE'

            target.install += [
                '$(PIP) install $(PIP_INSTALL_OPTIONS) -U "pip ~=10.0" wheel',
                '$(PIP) install $(PIP_INSTALL_OPTIONS) -U -r $(' + reqs_var + ')',
            ]
            target.deps += ['$(PYTHON_REQUIREMENTS_FILE)', '$(' + reqs_var + ')', 'setup.py']

    @subscribe('medikit.on_start', priority=ABSOLUTE_PRIORITY)
    def on_start(self, event):
        """
        **Events**

        - ``medikit.feature.python.on_generate`` (with the same ``ProjectEvent`` we got, todo: why not deprecate
          it in favor of higher priority medikit.on_start?)

        **Files**

        - ``<yourpackage>/__init__.py``
        - ``MANIFEST.in``
        - ``README.rst``
        - ``classifiers.txt``
        - ``requirements-dev.txt``
        - ``requirements.txt``
        - ``setup.cfg``
        - ``setup.py`` (overwritten)

        :param ProjectEvent event:
        """
        self.dispatcher.dispatch(__name__ + '.on_generate', event)

        # Our config object
        python_config = event.config['python']

        # Some metadata that will prove useful.
        self.render_empty_files('classifiers.txt', 'README.rst')
        self.render_file_inline('MANIFEST.in', 'include *.txt')
        self.render_file_inline(
            'setup.cfg', '''
                    [bdist_wheel]
                    # This flag says that the code is written to work on both Python 2 and Python
                    # 3. If at all possible, it is good practice to do this. If you cannot, you
                    # will need to generate wheels for each Python version that you support.
                    universal=1
                    
                    [metadata]
                    description-file = README.rst
                '''
        )

        # XXX there is an important side effect in get_init_files that defines namespace packages for later use in
        # setup.py rendering. Even if we do not want to create the packages here, it is important to make the call
        # anyway.
        for dirname, filename, context in python_config.get_init_files():
            if python_config.create_packages:
                if not os.path.exists(dirname):
                    os.makedirs(dirname)
                self.render_file(filename, 'python/package_init.py.j2', context)

        # render version file, try to import version from version.txt if available.
        try:
            with open('version.txt') as f:
                version = f.read().strip()
        except IOError:
            version = '0.0.0'
        self.render_file_inline(python_config.version_file, "__version__ = '{}'".format(version))

        setup = python_config.get_setup()

        context = {
            'url': setup.pop('url', 'http://example.com/'),
            'download_url': setup.pop('download_url', 'http://example.com/'),
        }

        for k, v in context.items():
            context[k] = context[k].format(
                name=setup['name'],
                user=getuser(),
                version='{version}',
            )

        context.update(
            {
                'entry_points': setup.pop('entry_points', {}),
                'extras_require': python_config.get('extras_require'),
                'install_requires': python_config.get('install_requires'),
                'python': python_config,
                'setup': setup,
                'banner': get_override_warning_banner(),
            }
        )

        # Render (with overwriting) the allmighty setup.py
        self.render_file('setup.py', 'python/setup.py.j2', context, override=True)

    @subscribe('medikit.on_end', priority=ABSOLUTE_PRIORITY)
    def on_end(self, event):
        # Our config object
        python_config = event.config['python']

        # Pip / PyPI
        pip_command = get_pip_command()
        pip_options, _ = pip_command.parse_args([])
        session = pip_command._build_session(pip_options)
        repository = PyPIRepository(pip_options, session)

        for extra in itertools.chain((None,), python_config.get_extras()):
            requirements_file = 'requirements{}.txt'.format('-' + extra if extra else '')

            if python_config.override_requirements or not os.path.exists(requirements_file):
                tmpfile = tempfile.NamedTemporaryFile(mode='wt', delete=False)
                if extra:
                    tmpfile.write('\n'.join(python_config.get_requirements(extra=extra)))
                else:
                    tmpfile.write('\n'.join(python_config.get_requirements()))
                tmpfile.flush()
                constraints = list(
                    parse_requirements(
                        tmpfile.name, finder=repository.finder, session=repository.session, options=pip_options
                    )
                )
                resolver = Resolver(constraints, repository, prereleases=False, clear_caches=False, allow_unsafe=False)

                self.render_file_inline(
                    requirements_file,
                    '\n'.join(
                        (
                            '-e .{}'.format('[' + extra + ']' if extra else ''),
                            *(('-r requirements.txt',) if extra else ()),
                            *sorted(
                                format_requirement(req) for req in resolver.resolve(max_rounds=10)
                                if req.name != python_config.get('name')
                            ),
                        )
                    ),
                    override=python_config.override_requirements
                )



"""
GNU Make support.

"""

import medikit
from medikit.events import subscribe
from medikit.feature import Feature, HIGH_PRIORITY
from medikit.feature.make.config import MakeConfig
from medikit.feature.make.events import MakefileEvent
from medikit.feature.make.resources import MakefileTarget, Makefile, InstallScript, CleanScript
from medikit.feature.make.utils import which
from medikit.structs import Script

__all__ = [
    'CleanScript',
    'InstallScript',
    'MakeConfig',
    'MakeFeature',
    'Makefile',
    'MakefileEvent',
    'MakefileTarget',
    'which',
]


class MakeFeature(Feature):
    Config = MakeConfig

    def configure(self):
        self.makefile = Makefile()

    @subscribe('medikit.on_start', priority=HIGH_PRIORITY)
    def on_start(self, event):
        """
        :param ProjectEvent event:
        """

        for k in event.variables:
            self.makefile[k.upper()] = event.variables[k]

        self.makefile.updateleft((
            'QUICK',
            '',
        ), )

        self.makefile.add_install_target()

        for extra in event.config['make'].extras:
            self.makefile.add_install_target(extra)

        self.makefile.add_target('quick', Script('@printf ""'), phony=True, hidden=True)

        self.dispatcher.dispatch(
            MakeConfig.on_generate, MakefileEvent(event.config.package_name, self.makefile, event.config)
        )

        if event.config['make'].include_medikit_targets:
            self.add_medikit_targets(event.config['make'])

        # Recipe courtesy of https://marmelab.com/blog/2016/02/29/auto-documented-makefile.html
        self.makefile.add_target(
            'help',
            r"""
            @echo "Available commands:"
            @echo
            @grep -E '^[a-zA-Z_-]+:.*?##[\s]?.*$$' $(MAKEFILE_LIST) | sort | awk 'BEGIN {FS = ":.*?##"}; {printf "    make \033[36m%-30s\033[0m %s\n", $$1, $$2}'
            @echo
            """,
            phony=True,
            doc='Shows available commands.'
        )

        # Actual rendering of the Makefile
        self.render_file_inline('Makefile', self.makefile.__str__(), override=True)

    def add_medikit_targets(self, config):
        if not 'PYTHON' in self.makefile:
            self.makefile['PYTHON'] = which('python')
        self.makefile['MEDIKIT'] = '$(PYTHON) -m medikit'
        self.makefile['MEDIKIT_UPDATE_OPTIONS'] = ''
        self.makefile['MEDIKIT_VERSION'] = medikit.__version__

        source = [
            'import medikit, pip, sys',
            'from packaging.version import Version',
            'sys.exit(0 if (Version(medikit.__version__) >= Version("$(MEDIKIT_VERSION)")) and (Version(pip.__version__) < Version("10")) else 1)',
        ]

        self.makefile.add_target(
            'medikit',
            '@$(PYTHON) -c {!r} || $(PYTHON) -m pip install -U "pip ~=10.0" "medikit>=$(MEDIKIT_VERSION)"'.format(
                '; '.join(source)
            ),
            phony=True,
            hidden=True,
            doc='Checks installed medikit version and updates it if it is outdated.'
        )

        self.makefile.add_target(
            'update',
            '$(MEDIKIT) update $(MEDIKIT_UPDATE_OPTIONS)',
            deps=('medikit', ),
            phony=True,
            doc='''Update project artifacts using medikit.'''
        )

        # TODO this should adapt to langauges included, and be removed if no language
        # For example, requirements*.txt are specific to python, using classic setuptools.
        self.makefile.add_target(
            'update-requirements',
            'MEDIKIT_UPDATE_OPTIONS="--override-requirements" $(MAKE) update',
            phony=True,
            doc='''Update project artifacts using medikit, including requirements files.'''
        )



from medikit.feature import Feature


class MakeConfig(Feature.Config):
    on_generate = 'medikit.feature.make.on_generate'
    """Happens during the makefile generation."""

    def __init__(self):
        self.include_medikit_targets = True
        self.extras = {'dev'}

    def disable_medikit_targets(self):
        self.include_medikit_targets = False


import itertools
import textwrap
from collections import deque, namedtuple

from medikit.structs import Script
from medikit.utils import get_override_warning_banner

MakefileTarget = namedtuple('MakefileTarget', ['deps', 'rule', 'doc'])


class Makefile(object):
    @property
    def targets(self):
        for key in self._target_order:
            yield key, self._target_values[key]

    @property
    def environ(self):
        return self._env_values

    def __init__(self):
        self._env_order, self._env_values, self._env_assignment_operators = deque(), {}, {}
        self._target_order, self._target_values = deque(), {}
        self.hidden = set()
        self.phony = set()

    def __delitem__(self, key):
        self._env_order.remove(key)
        del self._env_values[key]

    def __contains__(self, item):
        return item in self._env_values

    def __getitem__(self, item):
        return self._env_values[item]

    def __setitem__(self, key, value):
        self._env_values[key] = value
        if not key in self._env_order:
            self._env_order.append(key)

    def __iter__(self):
        for key in self._env_order:
            yield key, self._env_values[key]

    def __len__(self):
        return len(self._env_order)

    def __str__(self):
        content = [
            get_override_warning_banner(),
            '',
        ]

        if len(self):
            for k, v in self:
                v = textwrap.dedent(str(v)).strip()
                v = v.replace('\n', ' \\\n' + ' ' * (len(k) + 4))
                content.append('{} {} {}'.format(k, self._env_assignment_operators.get(k, '?='), v))
            content.append('')

        if len(self.phony):
            content.append('.PHONY: ' + ' '.join(sorted(self.phony)))
            content.append('')

        for target, details in self.targets:
            deps, rule, doc = details

            if hasattr(rule, 'render'):
                content += list(rule.render(target, deps, doc))
            else:
                content.append(
                    '{}: {}  {} {}'.format(
                        target, ' '.join(deps), '#' if target in self.hidden else '##',
                        doc.replace('\n', ' ') if doc else ''
                    ).strip()
                )

                script = textwrap.dedent(str(rule)).strip()

                for line in script.split('\n'):
                    content.append('\t' + line)

            content.append('')

        return '\n'.join(content)

    def keys(self):
        return list(self._env_order)

    def add_target(self, target, rule, *, deps=None, phony=False, first=False, doc=None, hidden=False):
        if target in self._target_order:
            raise RuntimeError('Duplicate definition for make target «{}».'.format(target))

        if isinstance(rule, str):
            rule = Script(rule)

        self._target_values[target] = MakefileTarget(
            deps=tuple(deps) if deps else tuple(),
            rule=rule,
            doc=textwrap.dedent(doc or '').strip(),
        )

        self._target_order.appendleft(target) if first else self._target_order.append(target)

        if phony:
            self.phony.add(target)

        if hidden:
            self.hidden.add(target)

    def get_clean_target(self):
        if not self.has_target('clean'):
            self.add_target('clean', CleanScript(), phony=True, doc='''Cleans up the working copy.''')
        return self.get_target('clean')

    def add_install_target(self, extra=None):
        if extra:
            target = 'install-' + extra
            doc = 'Installs the project (with ' + extra + ' dependencies).'
        else:
            target = 'install'
            doc = 'Installs the project.'

        if not self.has_target(target):
            self.add_target(target, InstallScript(), phony=True, doc=doc)

        clean_target = self.get_clean_target()
        marker = '.medikit/' + target
        if not marker in clean_target.remove:
            clean_target.remove.append(marker)
        return self.get_target(target)

    def get_target(self, target):
        return self._target_values[target][1]

    def has_target(self, target):
        return target in self._target_values

    def set_deps(self, target, deps=None):
        self._target_values[target] = (
            deps or list(),
            self._target_values[target][1],
            self._target_values[target][2],
        )

    def set_assignment_operator(self, key, value):
        assert value in ('?=', '=', '+=', ':=', '::=', '!='), 'Invalid operator'
        self._env_assignment_operators[key] = value

    def setleft(self, key, value):
        self._env_values[key] = value
        if key in self._env_order:
            self._env_order.remove(key)
        self._env_order.appendleft(key)

    def updateleft(self, *lst):
        for key, value in reversed(lst):
            self.setleft(key, value)


class InstallScript(Script):
    def __init__(self, script=None):
        super(InstallScript, self).__init__(script)

        self.before_install = []
        self.install = self.script
        self.after_install = []

        self.deps = []

    def __iter__(self):
        yield '@if [ -z "$(QUICK)" ]; then \\'
        for line in map(
            lambda x: '    {} ; \\'.format(x), itertools.chain(self.before_install, self.install, self.after_install)
        ):
            yield line
        yield 'fi'

    def render(self, name, deps, doc):
        tab = '\t'
        yield '{name}: .medikit/{name} {deps}  ## {doc}'.format(name=name, deps=' '.join(deps), doc=doc)
        yield '.medikit/{name}: {deps}'.format(name=name, deps=' '.join(sorted(set(self.deps))))
        yield tab + '$(eval target := $(shell echo $@ | rev | cut -d/ -f1 | rev))'
        yield 'ifeq ($(filter quick,$(MAKECMDGOALS)),quick)'
        yield tab + r'@printf "Skipping \033[36m%s\033[0m because of \033[36mquick\033[0m target.\n" $(target)'
        yield 'else ifneq ($(QUICK),)'
        yield tab + r'@printf "Skipping \033[36m%s\033[0m because \033[36m$$QUICK\033[0m is not empty.\n" $(target)'
        yield 'else'
        yield tab + r'@printf "Applying \033[36m%s\033[0m target...\n" $(target)'
        for line in itertools.chain(self.before_install, self.install, self.after_install):
            yield tab + line
        yield tab + '@mkdir -p .medikit; touch $@'
        yield 'endif'


class CleanScript(Script):
    # We should not clean .medikit subdirectories here, as it will deny releases.
    # TODO: move into python feature
    remove = [
        'build',
        'dist',
        '*.egg-info',
    ]

    def __iter__(self):
        # cleanup build directories
        yield 'rm -rf {}'.format(' '.join(self.remove))
        # cleanup python bytecode -
        yield 'find . -name __pycache__ -type d | xargs rm -rf'



from pip._vendor.distlib.util import parse_requirement

from medikit.events import subscribe
from . import Feature, SUPPORT_PRIORITY


class SphinxConfig(Feature.Config):
    def __init__(self):
        self._theme = None

    def set_theme(self, theme):
        """
        Sets the theme. Prefer using the .theme property.

        :param theme: Requirement for theme
        """
        self._theme = parse_requirement(theme)

    def get_theme(self):
        """
        Gets the theme. Prefer using the .theme property.

        :return: parsed requirement
        """
        return self._theme

    theme = property(
        fget=get_theme, fset=set_theme, doc="Sphinx theme to use, that should be parsable as a requirement."
    )


class SphinxFeature(Feature):
    Config = SphinxConfig

    @subscribe('medikit.feature.python.on_generate')
    def on_python_generate(self, event):
        event.config['python'].add_requirements(dev=[
            'sphinx ~=1.7',
        ])

        theme = event.config['sphinx'].get_theme()
        if theme:
            event.config['python'].add_requirements(dev=[theme.requirement])

    @subscribe('medikit.feature.make.on_generate', priority=SUPPORT_PRIORITY)
    def on_make_generate(self, event):
        makefile = event.makefile

        makefile['SPHINX_BUILD'] = '$(PYTHON_DIRNAME)/sphinx-build'
        makefile['SPHINX_OPTIONS'] = ''
        makefile['SPHINX_SOURCEDIR'] = 'docs'
        makefile['SPHINX_BUILDDIR'] = '$(SPHINX_SOURCEDIR)/_build'

        makefile.add_target(
            '$(SPHINX_SOURCEDIR)',
            '''
            $(SPHINX_BUILD) -b html -D latex_paper_size=a4 $(SPHINX_OPTIONS) $(SPHINX_SOURCEDIR) $(SPHINX_BUILDDIR)/html
        ''',
            deps=('install-dev', ),
            phony=True
        )
